import torch
from torch import Tensor
from torch.nn import Parameter
import numpy as np
from typing import List, Union, Literal
from numbers import Number


class _DiffBinning(torch.nn.Module):
    """Base class for differentiable binning layers.

    The binned layers are multiplexed as a new dimension in the output tensor.

    """
    def __init__(self, first_bin: Union[None, Number] = None, last_bin: Union[None, Number] = None,
                 nbins: Number = 10, bin_centers: Union[None, Tensor] = None, energy_ratio: Number = 1.) -> None:
        super().__init__()
        if bin_centers is not None:
            if first_bin is not None or last_bin is not None:
                raise ValueError("Cannot specify bin_centers and first_bin/last_bin")
            self.bin_centers: Parameter = Parameter(bin_centers, requires_grad=False)
            self.sigma: float = (bin_centers[1] - bin_centers[0]).item()
            self.er: float = energy_ratio
        else:
            if first_bin is None or last_bin is None:
                raise ValueError("Either bin_centers or first_bin and last_bin must be specified")
            self.bin_centers: Parameter = Parameter(torch.linspace(first_bin, last_bin, nbins), requires_grad=False)
            self.sigma: float = (last_bin - first_bin) / nbins
            self.er: float = energy_ratio

    def forward(self, x: Tensor) -> Tensor:
        ratio = 2.5066  # Manually aproximated this ratio so that the energy is 1
        x = torch.clamp(x, self.bin_centers[0], self.bin_centers[-1])
        if x.dim() == 4:
            bins = self.bin_centers[None, None, None, None, :]
            x = x[:, :, :, :, None]
            return (self.er / ratio) * torch.exp(-((x - bins) ** 2) / (2 * self.sigma ** 2))
        elif x.dim() == 3:
            bins = self.bin_centers[None, None, None, :]
            x = x[:, :, :, None]
            return (self.er / ratio) * torch.exp(-((x - bins) ** 2) / (2 * self.sigma ** 2))
        elif x.dim() == 2:
            bins = self.bin_centers[None, None, :]
            x = x[:, :, None]
            return (self.er / ratio) * torch.exp(-((x - bins) ** 2) / (2 * self.sigma ** 2))
        elif x.dim() == 1:
            bins = self.bin_centers[None, :]
            x = x[:, None]
            return (self.er / ratio) * torch.exp(-((x - bins) ** 2) / (2 * self.sigma ** 2))
        else:
            raise ValueError(f"Unsupported dim {x.dim()}")

    @property
    def bins(self) -> Tensor:
        return self.bin_centers


class DiffHistogram(_DiffBinning):
    def __init__(self, first_bin: Union[None, Number] = None, last_bin: Union[None, Number] = None,
                 nbins: Number = 10, bin_centers: Union[None, Tensor] = None, energy_ratio: Number = 1.,
                 pooling: Union[None, Literal['sum', 'avg'], torch.nn.Module] = None, keepdims: bool = True,
                 pool_dims: List[Number] = [2, 3]):
        super().__init__(first_bin=first_bin, last_bin=last_bin, nbins=nbins, bin_centers=bin_centers,
                         energy_ratio=energy_ratio)
        self.keepdims: bool = keepdims
        self.pool_dims: List[Number] = pool_dims
        self.pooling_method: str = pooling
        if pooling is None:
            self.pooling = lambda x: x  # Identity
        elif pooling == 'sum':
            self.pooling = lambda x: x.sum(dim=pool_dims, keepdims=keepdims)
        elif pooling == 'avg':
            self.pooling = lambda x: x.mean(dim=pool_dims, keepdims=keepdims)
        elif isinstance(pooling, torch.nn.Module):
            self.pooling = pooling
            self.pooling_method = f"custom_{pooling.__class__.__name__}"
        else:
            raise ValueError(f"Unknown pooling method {pooling}")

    def forward(self, x: Tensor) -> Tensor:
        if x.dim() != 4:
            raise ValueError("Input must be a BxCxHxW tensor")
        per_channel = []
        for ch in range(0, x.size(1)):
            b_h_w_bin = super().forward(x[:, ch:ch+1, :, :])[:, 0, :, :, :]  # (B, H, W, BIN)
            b_bin_h_w = b_h_w_bin.permute(0, 3, 1, 2)  # (B, BIN, H, W)
            per_channel.append(b_bin_h_w)
        x = torch.cat(per_channel, dim=1)  # (B, CxBIN, H, W)
        return self.pooling(x)


def get_intraclass_variances(hist: Tensor, levels: Tensor, e=1e-10) -> Tensor:
    """Calculate the weighted intraclass variances.

    Args:
        hist (Tensor): A tensor who has histograms along a dimension.
        levels (Tensor): A tensor with the values the histogram bins. They are assumed to be equidistant.
        e (float, optional): A small value to avoid division by zero. Defaults to 1e-10.

    Returns:
        Tensor: The tensor containing the weighted intraclass variances.
    """
    h_sum = hist.sum(dim=-1, keepdim=True)
    low_coef = hist.cumsum(dim=-1) / h_sum
    low_E_x = ((levels * hist).cumsum(dim=-1) / (hist.cumsum(dim=-1) + e))
    low_var = (((levels - low_E_x) ** 2) * hist).cumsum(dim=-1) / (hist.cumsum(dim=-1) + e)

    f_levels = torch.flip(levels, dims=[0])
    f_hist = torch.flip(hist, dims=[-1])
    f_high_E_x = ((f_levels * f_hist).cumsum(dim=-1) / (f_hist.cumsum(dim=-1) + e))
    f_high_var = (((f_levels - f_high_E_x) ** 2) * f_hist).cumsum(dim=-1) / (f_hist.cumsum(dim=-1) + e)
    high_var = torch.flip(f_high_var, dims=[-1])

    weighted_intraclass_variances = low_var * low_coef + high_var * (1 - low_coef)
    return weighted_intraclass_variances


def hard_otsu_threshold(histograms: Tensor, dim: int = 1, e=1e-10) -> Tensor:
    """
    Applies the hard Otsu thresholding algorithm to the input histograms.

    Args:
        histograms (Tensor): The input histograms. It can be a 1D or higher-dimensional tensor.
        dim (int, optional): The dimension along which the histograms are computed. Default is 1.
        e (float, optional): A small value added to the denominator to avoid division by zero. Default is 1e-10.

    Returns:
        Tensor: An integer tensor with the histogram dimension beeing the index of the Otsu Threhold for the
            respective histogram.

    Raises:
        None

    Examples:
        >>> histograms = torch.tensor([0.7, 0.9, 0.1, 0.6])
        >>> thresholded = hard_otsu_threshold(histograms)
        >>> print(thresholded)
        tensor([2])
    """
    with torch.no_grad():
        device = histograms.device
    if histograms.dim() == 1 and dim != 0:
        histograms = histograms[None, :]
    nb_bins = histograms.size(dim)
    ld_histograms = histograms.swapaxes(dim, -1)  # moving to (..., BIN)
    levels = torch.arange(0, nb_bins, dtype=torch.float, device=device).view(*([1] * (histograms.dim()-1)), -1)
    for chunc_n in range(0, ld_histograms.size(-1)//nb_bins):
        hist = ld_histograms[..., chunc_n*nb_bins:(chunc_n+1)*nb_bins]
        intraclass_variances = get_intraclass_variances(hist, levels=levels, e=e)
    ld_res = torch.argmin(intraclass_variances, dim=-1, keepdim=True)
    return ld_res.swapaxes(-1, dim)


def diff_otsu_threshold(histograms: Tensor, dim: int = 1, t=.01, nb_bins: int = -1, e=1e-10) -> Tensor:
    """
    Compute the differential Otsu threshold for a given histogram.

    Args:
        histograms (Tensor): The input histogram(s) tensor.
        dim (int, optional): The dimension along which the histogram(s) are computed. Defaults to 1.
        t (float, optional): The temperature parameter for softmax. Defaults to 0.01.
        nb_bins (int, optional): The number of bins in the histogram. If -1, it is inferred from the input tensor.
            Defaults to -1.
        e (float, optional): A small value added to the denominator for numerical stability. Defaults to 1e-10.

    Returns:
        Tensor: The differential Otsu threshold tensor.

    Raises:
        ValueError: If the input histogram tensor is not 1-dimensional and `dim` is not 0.

    Examples:
        >>> histograms = torch.rand(3, 128, 100, 101)
        >>> diff_otsu_threshold(histograms).size()
        torch.Size([3, 128, 100, 101])
        >>> torch.torch.aminmax(diff_otsu_threshold(histograms))
        (tensor(0.00...), tensor(.99....))
    """
    device = histograms.device
    if histograms.dim() == 1 and dim != 0:
        histograms = histograms[None, :]
    if nb_bins < 0:
        nb_bins = histograms.size(dim)
    ld_histograms = histograms.swapaxes(dim, -1)
    levels = torch.arange(0, nb_bins, dtype=torch.float, device=device).view(*([1] * (histograms.dim()-1)), -1)
    ld_otsu_chuncs = []
    for chunc_n in range(0, ld_histograms.size(-1)//nb_bins):
        hist = ld_histograms[..., chunc_n*nb_bins:(chunc_n+1)*nb_bins]
        intraclass_variances = get_intraclass_variances(hist, levels=levels, e=e)
        #  Bringing the intraclass variances to a standard normal distribution for a better softmax
        intraclass_variances = intraclass_variances - intraclass_variances.mean(dim=-1, keepdim=True)
        intraclass_variances = intraclass_variances / (intraclass_variances.std(dim=-1, keepdim=True) + e)
        optimal_threshold_prob = (-intraclass_variances / t).softmax(dim=-1)
        ld_otsu_chuncs.append(optimal_threshold_prob)
    ld_otsu_chuncs = torch.cat(ld_otsu_chuncs, dim=-1)
    return ld_otsu_chuncs.swapaxes(-1, dim)


# def otsu_threshold(histograms: Tensor, dim: int = 1, t=.01, nb_bins: int = -1, e=1e-10) -> Tensor:
#     device = histograms.device
#     if histograms.dim() == 1 and dim != 0:
#         histograms = histograms[None, :]
#     if nb_bins < 0:
#         nb_bins = histograms.size(dim)
#     ld_histograms = histograms.swapaxes(dim, -1)
#     levels = torch.arange(0, nb_bins, dtype=torch.float, device=device).view(*([1] * (histograms.dim()-1)), -1)
#     ld_otsu_chuncs = []
#     for chunc_n in range(0, ld_histograms.size(-1)//nb_bins):
#         hist = ld_histograms[..., chunc_n*nb_bins:(chunc_n+1)*nb_bins]
#         h_sum = hist.sum(dim=-1, keepdim=True)
#         low_coef = hist.cumsum(dim=-1)/h_sum
#         high_coef = 1 - low_coef
#         low_E_x = ((levels * hist).cumsum(dim=-1) / (hist.cumsum(dim=-1) + e))
#         low_var = (((levels - low_E_x) ** 2) * hist).cumsum(dim=-1) / (hist.cumsum(dim=-1) + e)

#         f_levels = torch.flip(levels, dims=[0])
#         f_hist = torch.flip(hist, dims=[-1])
#         f_high_E_x = ((f_levels * f_hist).cumsum(dim=-1) / (f_hist.cumsum(dim=-1) + e))
#         f_high_var = (((f_levels - f_high_E_x) ** 2) * f_hist).cumsum(dim=-1) / (f_hist.cumsum(dim=-1) + e)
#         high_var = torch.flip(f_high_var, dims=[-1])

#         intraclass_variances = low_var * low_coef + high_var * high_coef
#         intraclass_variances = intraclass_variances - intraclass_variances.mean(dim=-1, keepdim=True)
#         intraclass_variances = intraclass_variances / (intraclass_variances.std(dim=-1, keepdim=True) + e)
#         optimal_threshold_prob = (-intraclass_variances / t).softmax(dim=-1)
#         ld_otsu_chuncs.append(optimal_threshold_prob)
#     ld_otsu_chuncs = torch.cat(ld_otsu_chuncs, dim=-1)
#     return ld_otsu_chuncs.swapaxes(-1, dim)


class DiffOtsu(torch.nn.Module):
    def __init__(self, first_bin: Union[None, Number] = None, last_bin: Union[None, Number] = None,
                 nbins: Number = 10, bin_centers: Union[None, Tensor] = None, energy_ratio: Number = 1.,
                 temperature: float = .01, e: float = 1e-10) -> None:
        super().__init__()
        self.nbins: int = nbins
        self.t: float = temperature
        self.e: float = e
        self.binning_layer = DiffHistogram(first_bin=first_bin, last_bin=last_bin, nbins=nbins,
                                           bin_centers=bin_centers, energy_ratio=energy_ratio, pooling=None,
                                           keepdims=True)

    def forward(self, x: Tensor) -> Tensor:
        if x.dim() != 4:
            raise ValueError("Input must be a BxCxHxW tensor")
        if x.size(1) != 1:
            # TODO: Implement (cumsum that resets every bins)
            raise NotImplementedError("Number of bins must match the number of channels")
        bins = self.binning_layer.bins[None, :, None, None]
        x = torch.clamp(x, bins.min(), bins.max())
        binned_x = self.binning_layer(x)  # (B, 1xBIN, H, W)
        histogram = binned_x.sum(dim=(2, 3), keepdims=True)  # (B, 1xBIN, 1, 1)
        thr_prob = diff_otsu_threshold(histogram, dim=1, t=self.t, nb_bins=self.nbins, e=self.e)
        result_chunks = []
        for chunk_n in range(0, binned_x.size(1) // self.nbins):
            chunk_start = chunk_n * self.nbins
            chunk_end = (chunk_n + 1) * self.nbins
            binned_chunk = binned_x[:, chunk_start: chunk_end, :, :]
            chunk_cum_prob = thr_prob[:, chunk_start: chunk_end, :, :].cumsum(dim=1)
            result_chunks.append((binned_chunk * chunk_cum_prob).sum(dim=1, keepdims=True))
        return torch.cat(result_chunks, dim=1)


def __avg_pool(x, output_size) -> torch.Tensor:
    return torch.nn.functional.adaptive_avg_pool2d(x, (output_size, output_size))


class DiffHOG(torch.nn.Module):
    def __init__(self, cell_size: Number = 8,
                 pooling: Union[Literal['avg'], Literal['sum'], None, torch.nn.Module] = None,
                 keepdims: bool = True, nbins: int = 9, signed_gradient: bool = False) -> None:
        super().__init__()
        self.cell_size: Number = cell_size
        self.keepdims: bool = keepdims
        self.nbins: int = nbins
        self.signed_gradient: bool = signed_gradient
        angle_bins = torch.linspace(0, 3.14159265359 * (1+signed_gradient), nbins + 1, requires_grad=False)
        angle_bins = (angle_bins[:-1] + angle_bins[1:]) / 2
        self.histogram_layer = DiffHistogram(bin_centers=angle_bins, energy_ratio=1, pooling='None')
        self.grad_filter = torch.nn.Conv2d(1, 2, 3, padding=1, bias=False)
        self.grad_filter.weight[0, 0, :, 0] = [1, 2, 1]
        self.grad_filter.weight[0, 0, :, 2] = [-1, -2, -1]
        self.grad_filter.weight[0, 0, 0, :] = [1, 2, 1]
        self.grad_filter.weight[0, 0, 2, :] = [-1, -2, -1]
        self.pooling_method: str = pooling

        if pooling is None:
            self.pooling = lambda x: x

        elif pooling == 'avg' and cell_size < 0:
            self.pooling = lambda x: x.mean(dim=(2, 3), keepdims=self.keepdims)
        elif pooling == 'avg' and cell_size > 1:
            if int(cell_size) != cell_size:
                raise ValueError("Cell size must be an integer when > 1")
            self.pooling = lambda x: torch.nn.functional.avg_pool2d(x, cell_size, stride=cell_size)
        elif pooling == 'avg' and 0. < cell_size < 1.:
            if (int(1/cell_size) - 1/cell_size) ** 2 > 1e-6:
                raise ValueError("1/cell_size must be an integer when 0 < cell_size < 1")
            else:
                output_size = int(np.round(1/cell_size))
            self.pooling = lambda x: __avg_pool(x, output_size)

        elif pooling == 'sum' and cell_size < 0:
            self.pooling = lambda x: x.sum(dim=(2, 3), keepdims=self.keepdims)
        elif pooling == 'sum' and cell_size > 1:
            if int(cell_size) != cell_size:
                raise ValueError("Cell size must be an integer when > 1")
            self.pooling = lambda x: torch.nn.functional.avg_pool2d(x, cell_size, stride=cell_size)
        elif pooling == 'sum' and 0. < cell_size < 1.:
            if (int(1/cell_size) - 1/cell_size) ** 2 > 1e-6:
                raise ValueError("1/cell_size must be an integer when 0 < cell_size < 1")
            else:
                output_size = int(np.round(1/cell_size))
            self.pooling = lambda x: __avg_pool(x) * output_size ** 2
        elif isinstance(pooling, torch.nn.Module):
            self.pooling = pooling
            self.pooling_method = f"custom_{pooling.__class__.__name__}"
        else:
            raise ValueError(f"Unknown pooling method {pooling}")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if x.dim() != 4:
            raise ValueError("Input must be a BxCxHxW tensor")
        per_channel = []
        for ch in range(0, x.size(1)):
            ch_x = self.grad_filter(x[:, ch:ch+1, :, :])  # (B, 2, H, W)
            if not self.signed_gradient:
                ch_x = ch_x * torch.tanh(ch_x*10000)  # Approximate abs but differentiable
            magnitude = torch.sqrt(ch_x[:, 0] ** 2 + ch_x[:, 1] ** 2)[:, None, :, :]  # (B, 1, H, W)
            angles = torch.atan2(ch_x[:, 1], ch_x[:, 0])[:, None, :, :]  # (B, 1, H, W)
            binned_angles = self.histogram_layer(angles)  # (B, 1xBIN, H, W)
            binned_magnitudes = binned_angles * magnitude  # (B, 1xBIN, H, W)
            per_channel.append(binned_magnitudes)
        x = torch.cat(per_channel, dim=1)
        del per_channel
        return self.pooling(x)

    @property
    def angle_bins(self) -> torch.Tensor:
        return self.histogram_layer.bins



import torch
import pytest
from ptlbp import PointSampler, BinaryToOnehotChannelEmbeddings as BinaryToOnehotChannel
from skimage.feature import local_binary_pattern


def test_BinaryToOnehotChannel_encoding():
    bchw = torch.arange(210).view(1, 3, 7, 10)
    sz, bc = BinaryToOnehotChannel.img2embedding(bchw)
    new_bchw = BinaryToOnehotChannel.embedding2img(sz, bc)
    assert new_bchw.size() == bchw.size()
    assert torch.allclose(bchw, new_bchw)


@pytest.mark.parametrize("input_size", [1, 2, 3, 4, 8])
def test_BinaryToOnehotChannel_weightsize(input_size):
    oh = BinaryToOnehotChannel(in_channels=input_size)
    assert oh.linear.weight.size(0) == 2 ** input_size
    assert oh.linear.weight.size(1) == input_size
    assert len(oh.linear.weight.size()) == 2


@pytest.mark.parametrize("vector, maxpos", [[torch.tensor([[0, 0, 0]]), 0],
                                            [torch.tensor([[1, 0, 0]]), 1],
                                            [torch.tensor([[0, 1, 0]]), 2],
                                            [torch.tensor([[1, 1, 0]]), 3],
                                            [torch.tensor([[0, 0, 1]]), 4],
                                            [torch.tensor([[1, 0, 1]]), 5],
                                            [torch.tensor([[0, 1, 1]]), 6],
                                            [torch.tensor([[1, 1, 1]]), 7]])
def test_BinaryToOnehotChannel_powerset(vector, maxpos):
    """Test the powerset matrix generator BinaryToOnehotChannel is correct in the case of 3 bits"""
    W = torch.tensor(BinaryToOnehotChannel.powerset_map(3), dtype=torch.float)
    vector = vector.float()
    output = 1 + (vector @ W.T) - ((1-vector) @ W.T)
    assert list(output.size()) == [1, 8]
    assert torch.argmax(output[0, :]).item() == maxpos


img1 = torch.zeros([1, 3, 10, 11], dtype=torch.float32)
img2 = torch.zeros([1, 3, 10, 11], dtype=torch.float32)
img2[0, 0, :, :] = 1
img3 = torch.zeros([1, 3, 10, 11], dtype=torch.float32)
img3[0, :2, :, :] = 1
img4 = torch.zeros([1, 3, 10, 11], dtype=torch.float32)
img4[0, :, :, :] = 1


@pytest.mark.parametrize("input_img, active_output_channel", [[img1[:, :, :1, :1], 0], [img2[:, :, :1, :1], 1],
                                                              [img3[:, :, :1, :1], 3], [img4[:, :, :1, :1], 7]])
def test_BinaryToOnehotChannel_single_pixel(input_img, active_output_channel):
    """Test if the argmax of the output BinaryToOnehotChannel applied on a single pixel image is the same as the expected output"""
    oh = BinaryToOnehotChannel(in_channels=3)
    output = oh(input_img)
    histogram = output[0, :, 0, 0]
    assert torch.argmax(histogram).item() == active_output_channel


@pytest.mark.parametrize("input_img, active_output_channel", [[img1, 0], [img2, 1], [img3, 3], [img4, 7]])
def test_BinaryToOnehotChannel_pooled(input_img, active_output_channel):
    """Test if the argmax of the output BinaryToOnehotChannel aplied on images with uniform pixels per channel is the 
same as the expected output"""
    oh = BinaryToOnehotChannel(in_channels=3)
    output = oh(input_img)
    histogram = output.sum(dim=2).sum(dim=2)[0, :]
    assert torch.argmax(histogram).item() == active_output_channel


#  TODO(anguelos) find the corespondence between SCIKIT and SRS labels
def test_8bit_patterns():
    ps = PointSampler(1, "8r1")
    oh = BinaryToOnehotChannel(in_channels=8)

    #  This is specifically for 8r1 on a 5x5 image
    def get_center_pixel_pattern(img):
        return torch.argmax(oh(ps(img))[:, :, 1, 1], dim=1).item()

    def get_center_pixel_pattern_skimage(img):
        return local_binary_pattern(img[0, 0, :, :].numpy().astype("uint8"), 8, 1, method='default')[2, 2]

    img = torch.zeros([1, 1, 5, 5], dtype=torch.float32)
    assert get_center_pixel_pattern(img) == 0
    #  assert get_center_pixel_pattern(img) == get_center_pixel_pattern_skimage(img)

    img = torch.zeros([1, 1, 5, 5], dtype=torch.float32)
    img[0, 0, 2, 2] = 1
    assert get_center_pixel_pattern(img) == 0
    assert get_center_pixel_pattern(img) == get_center_pixel_pattern_skimage(img)

    img = torch.ones([1, 1, 5, 5], dtype=torch.float32)
    img[0, 0, 2, 2] = 0
    assert get_center_pixel_pattern(img) == 255
    assert get_center_pixel_pattern(img) == get_center_pixel_pattern_skimage(img)

    img = torch.zeros([1, 1, 5, 5], dtype=torch.float32)
    img[0, 0, 2:, :] = 1
    #  print(local_binary_pattern(img[0, 0, :, :].numpy().astype("uint8"), 8, 1, method='default'))
    assert get_center_pixel_pattern(img) == 199
    #  assert get_center_pixel_pattern(img) == get_center_pixel_pattern_skimage(img)



import torch


class CDistLoss(torch.nn.Module):
    def __init__(self, p=1., hardness_ratio=.1, cast_to_double=False, margin=0.0) -> None:
        super().__init__()
        self.p = p
        self.hardness_ratio = hardness_ratio
        self.cast_to_double = cast_to_double
        self.margin = margin

    def forward(self, x, y):
        assert x.size(0) == y.size(0)
        assert x.dim() == 2
        assert y.dim() == 1
        assert y.dtype == torch.int64 or y.dtype == torch.bool

        # TODO(anguelos) Verify this is the apropriate way to avoid a loss that is detached from the features
        if y.float().std() == 0:
            return .00000001 * torch.sum(x, dim=1)

        if self.cast_to_double:
            x = x.double()

        D = torch.cdist(x, x, p=self.p)

        with torch.no_grad():
            agree = (y[None, :] == y[:, None]).float()
            disagree = 1 - agree
            valid_rows = (agree.sum(dim=1) > 1) & (disagree.sum(dim=1) > 0)
            agree = agree[valid_rows, :]
            disagree = disagree[valid_rows, :]

        D = D[valid_rows, :]

        if D.size(0) == 0:
            return torch.tensor(0.0, device=D.device)

        with torch.no_grad():
            col_idx = torch.argsort(D, dim=1)
            col_idx = col_idx[:, 1:]  # Remove self match
            row_idx = torch.arange(D.size(0), dtype=torch.int64)[:, None].expand_as(col_idx)
            agree = agree[row_idx, col_idx]
            disagree = disagree[row_idx, col_idx]
            #valid_rows = (agree.sum(dim=1) > 0) & (disagree.sum(dim=1) > 0)

        dist = D[row_idx, col_idx]

        with torch.no_grad():
            found_agree = agree.cumsum(dim=1)
            found_agree /= (found_agree.sum(dim=1)[:, None] + .0000001)
            found_disagree = disagree.cumsum(dim=1)
            found_disagree /= (found_disagree.sum(dim=1)[:, None] + .0000001)
            negative_hardness = disagree * (found_disagree / (found_agree + self.hardness_ratio))
            positive_hardness = agree * ((found_disagree + self.hardness_ratio) / (found_agree + self.hardness_ratio))

        per_sample_score = torch.relu(dist * positive_hardness + self.margin - dist * negative_hardness).mean(dim=1)

        with torch.no_grad():
            sample_performance = (found_agree - found_disagree).sum(dim=1)
            sample_performance -= sample_performance.min()
            sample_weight = 1 / (sample_performance + 1)  # TODO(anguelos) test sample_weight = sample_performance + 1

        return (per_sample_score * sample_weight)


class CDistRatioLoss(torch.nn.Module):
    def __init__(self, p=1., hardness_ratio=.1, cast_to_double=False, margin=0.0) -> None:
        super().__init__()
        self.p = p
        self.hardness_ratio = hardness_ratio
        self.cast_to_double = cast_to_double
        self.margin = margin

    def forward(self, x, y):
        assert x.size(0) == y.size(0)
        assert x.dim() == 2
        assert y.dim() == 1
        assert y.dtype == torch.int64 or y.dtype == torch.bool

        # TODO(anguelos) Verify this is the apropriate way to avoid a loss that is detached from the features
        if y.float().std() == 0:
            return .00000001 * torch.sum(x, dim=1)

        if self.cast_to_double:
            x = x.double()

        D = torch.cdist(x, x, p=self.p)

        with torch.no_grad():
            agree = (y[None, :] == y[:, None]).float()
            disagree = 1 - agree
            valid_rows = (agree.sum(dim=1) > 1) & (disagree.sum(dim=1) > 0)
            agree = agree[valid_rows, :]
            disagree = disagree[valid_rows, :]

        D = D[valid_rows, :]

        if D.size(0) == 0:
            return torch.tensor(0.0, device=D.device)

        with torch.no_grad():
            col_idx = torch.argsort(D, dim=1)
            col_idx = col_idx[:, 1:]  # Remove self match
            row_idx = torch.arange(D.size(0), dtype=torch.int64)[:, None].expand_as(col_idx)
            agree = agree[row_idx, col_idx]
            disagree = disagree[row_idx, col_idx]

        dist = D[row_idx, col_idx]

        with torch.no_grad():
            agree_count = agree.sum(dim=1, keepdim=True)
            found_disagree = disagree.cumsum(dim=1)
            hardest_negatives = (found_disagree <= agree_count).float() * disagree

        d_p = (dist * agree).sum(dim=1)
        d_n = (dist * hardest_negatives).sum(dim=1)
        return (d_p / (d_p + d_n + .0001)) ** 2


class CDistTripletLoss(torch.nn.Module):
    def __init__(self, p=1., hardness_ratio=.1, cast_to_double=False, margin=0.0) -> None:
        super().__init__()
        self.p = p
        self.hardness_ratio = hardness_ratio
        self.cast_to_double = cast_to_double
        self.margin = margin

    def forward(self, x, y):
        assert x.size(0) == y.size(0)
        assert x.dim() == 2
        assert y.dim() == 1
        assert y.dtype == torch.int64 or y.dtype == torch.bool

        # TODO(anguelos) Verify this is the apropriate way to avoid a loss that is detached from the features
        if y.float().std() == 0:
            return .00000001 * torch.sum(x, dim=1)

        if self.cast_to_double:
            x = x.double()

        D = torch.cdist(x, x, p=self.p)

        with torch.no_grad():
            agree = (y[None, :] == y[:, None]).float()
            disagree = 1 - agree
            valid_rows = (agree.sum(dim=1) > 1) & (disagree.sum(dim=1) > 0)
            agree = agree[valid_rows, :]
            disagree = disagree[valid_rows, :]

        D = D[valid_rows, :]

        if D.size(0) == 0:
            return torch.tensor(0.0, device=D.device)

        with torch.no_grad():
            col_idx = torch.argsort(D, dim=1)
            col_idx = col_idx[:, 1:]  # Remove self match
            row_idx = torch.arange(D.size(0), dtype=torch.int64)[:, None].expand_as(col_idx)
            agree = agree[row_idx, col_idx]
            disagree = disagree[row_idx, col_idx]

        dist = D[row_idx, col_idx]

        with torch.no_grad():
            agree_count = agree.sum(dim=1, keepdim=True)
            found_disagree = disagree.cumsum(dim=1)
            hardest_negatives = (found_disagree <= agree_count).float() * disagree

        Dw_ap = (dist * agree).sum(dim=1)
        Dw_an = (dist * hardest_negatives).sum(dim=1)
        return torch.relu(Dw_ap - Dw_an + self.margin)


class CDistContrastiveLoss(torch.nn.Module):
    def __init__(self, p=1., hardness_ratio=.1, cast_to_double=False, margin=0.0) -> None:
        super().__init__()
        self.p = p
        self.hardness_ratio = hardness_ratio
        self.cast_to_double = cast_to_double
        self.margin = margin

    def forward(self, x, y):
        assert x.size(0) == y.size(0)
        assert x.dim() == 2
        assert y.dim() == 1
        assert y.dtype == torch.int64 or y.dtype == torch.bool

        # TODO(anguelos) Verify this is the apropriate way to avoid a loss that is detached from the features
        if y.float().std() == 0:
            return .00000001 * torch.sum(x, dim=1)

        if self.cast_to_double:
            x = x.double()

        D = torch.cdist(x, x, p=self.p)

        with torch.no_grad():
            agree = (y[None, :] == y[:, None]).float()
            disagree = 1 - agree
            valid_rows = (agree.sum(dim=1) > 1) & (disagree.sum(dim=1) > 0)
            agree = agree[valid_rows, :]
            disagree = disagree[valid_rows, :]

        D = D[valid_rows, :]

        if D.size(0) == 0:
            return torch.tensor(0.0, device=D.device)

        with torch.no_grad():
            col_idx = torch.argsort(D, dim=1)
            col_idx = col_idx[:, 1:]  # Remove self match
            row_idx = torch.arange(D.size(0), dtype=torch.int64)[:, None].expand_as(col_idx)
            agree = agree[row_idx, col_idx]
            disagree = disagree[row_idx, col_idx]

        dist = D[row_idx, col_idx]

        with torch.no_grad():
            agree_count = agree.sum(dim=1, keepdim=True)
            found_disagree = disagree.cumsum(dim=1)
            hardest_negatives = (found_disagree <= agree_count).float() * disagree

        return .5 * agree * dist + hardest_negatives * torch.relu(self.margin - dist)
